#include "geometry.h"
#include "smp_func.h"

/*********************************       Ray   ****************************/
bool Ray::OnRay(const Vector& pt) const
{
  Vector gap = pt - origin;
  return (fabs(Sin(gap.Angle() - angle) * gap.mod()) < FLOAT_EPS)
    ? true : false;
}

float Ray::Allowed_AngleDiff(float distance) const{
	return 20.0f;	//simplified by yjy ( for it's enough to use)
	//return float(10 - 5*exp(-distance * 0.2f));
}

float Ray::Allowed_AngleDiff(const Vector& point) const{
	return Allowed_AngleDiff(point.dist(origin));
}

bool Ray::InRightDir(const Vector& point) const{
	return fabs(NormalizeAngle((point - origin).Angle() - angle)) < Allowed_AngleDiff(point);
}

bool Ray::InRightDir(float ang) const{
	return fabs(NormalizeAngle(ang - angle)) < Allowed_AngleDiff();
}

bool Ray::InOppositeDir(const Vector& point) const{
	return fabs(NormalizeAngle((point - origin).Angle() - angle + 180)) < Allowed_AngleDiff(point);
}

bool Ray::InOppositeDir(float ang) const{
	return fabs(NormalizeAngle(ang - angle + 180)) < Allowed_AngleDiff();
}

bool Ray::intersection(const Line& l, Vector &pPt) const
{
  return l.RayIntersection(*this, pPt);
}

bool Ray::intersection(const Ray& r, Vector &pPt) const
{
	Line thisLine(*this), argLine(r);

	if (thisLine.SameSlope(argLine))
		return false;
	pPt = thisLine.intersection(argLine);

   //now make sure that the intersection is the correct direction on both lines 
	return  InRightDir(pPt) && r.InRightDir(pPt);
}

float Ray::intersection(const Line& l) const{
	return (-l.C -l.A * origin.x - l.B * origin.y) / ( l.A* direction.x + l.B * direction.y);
}
/* intersects a ray and a cricle */
/* return the number of solutions */
/* psol1 1 is not as far along the ray as psol2 */
int Ray::CircleIntersect(float radius, const Vector& center, Vector& psol1, Vector& psol2) const
{
	float a,b,c,disc;
	float t1, t2;
	a = Sqr(direction.x) + Sqr(direction.y);
	b = 2.0f * ((origin.x-center.x) * direction.x + (origin.y-center.y) * direction.y);
	c = Sqr(origin.x-center.x) + Sqr(origin.y-center.y) - Sqr(radius);

	disc = Sqr(b) - 4 * a * c;
	if (disc < 0) {
		return 0; // No solutions
	}
  
	disc = (float) sqrt(disc);
	t1 = (-b + disc) / (2.0f * a);
	t2 = (-b - disc) / (2.0f * a);
  
	if (t1 > t2) {
		float temp = t1;
		t1 = t2;
		t2 = temp;
	}
 
	if (t1 > 0.0) {
		if (t2 > 0.0) {
			psol1 = origin + direction * t1;
			psol2 = origin + direction * t2;
			return 2; // Two solutions
		}
		else{
			return 0;
		}
	}
	else if (t2 > 0.0) {
		psol1 = origin + direction * t2;
		psol2 = psol1;
		return 1;
	}
	else
		return 0;
	return 0;
}

Vector Ray::GetClosestPoint(const Vector& point) const{
  Line l(*this);
  Vector close_point = l.ProjectPoint(point);
  if (OnRay(close_point))
    return close_point;
  else
    return origin;
}

float Ray::StrechAngle(const Vector& point) const{
	return (float)fabs(NormalizeAngle((point - origin).Angle() - angle));
}

float Ray::GetClosestPointParam(const Vector& point) const{
	return DistanceFromOrigin(point) * Cos(StrechAngle(point));
}

float Ray::DistanceFromOrigin(const Vector& point) const{
	return (point -origin).mod();
}

void Ray::Normalize(){
	if (direction.mod2() != 0){
		direction /= direction.mod();
	}
}
/****************************   Line           ***************************/
Line::Line(float A, float B, float C) {
	this->A = A;
	this->B = B;
	this->C = C;
}

/********************************************************************************/

void Line::LineFromTwoPoints(const Vector& pt1, const Vector& pt2)
{
	float temp = (pt2.x - pt1.x);
	if (fabs(temp) < FLOAT_EPS) {
		//if (fabs(pt2.y - pt1.y) < FLOAT_EPS) ;
		A = 1;
		B = 0;
	}
	else {
		float m = (pt2.y - pt1.y) / temp;
		A = -m;
		B = 1;
	}  
	C = -(A * pt2.x + B * pt2.y);
}

/********************************************************************************/

void Line::LineFromRay(const Ray& r)   
{
	//if (fabs(r.direction.y) < FLOAT_EPS && fabs(r.direction.x) < FLOAT_EPS)

	LineFromTwoPoints(r.origin, r.origin + r.direction);
}

/********************************************************************************/

bool Line::PointOnLine(float x, float y) const{
  return (bool)(fabs(A * x + B * y + C) < FLOAT_EPS);
}

/********************************************************************************/

float Line::dist(const Vector& point) const{
	return (float)fabs( (A*point.x + B*point.y + C) / sqrt(Sqr(A) + Sqr(B)) );
}

float Line::dist2(const Vector& point) const{
	return (float) fabs( Sqr(A*point.x + B*point.y + C) / (Sqr(A) + Sqr(B)) );
}

float Line::angle() const{
	return ATan(-A / B);
}

/********************************************************************************/

float Line::get_y(float x) const{
	if ( B != 0 ) 
		return (-A*x - C)/B;

	return 0;
}

float Line::get_x(float y) const{
	if ( A != 0 ) 
		return (-B*y - C)/A;

	return 0;
}

/********************************************************************************/

bool Line::InBetween(const Vector& point, const Vector& end1, const Vector& end2) const{
//	if (!OnLine(end1) || !OnLine(end2))

	Vector p = ProjectPoint(point);
	float dist2 = end1.dist2(end2);
  
	return p.dist2(end1) <= dist2 && p.dist2(end2) <= dist2;
}

/********************************************************************************/

Vector Line::GetClosestPtInBetween(const Vector& pt, const Vector& end1, const Vector& end2) const{
	//if (!OnLine(end1) || !OnLine(end2))

	if (InBetween(pt, end1, end2))
		return ProjectPoint(pt);
	if (end1.dist2(pt) < end2.dist2(pt))
		return end1;
	else
		return end2;
}


/********************************************************************************/

Vector Line::intersection(const Line& l) const{
	Vector result = 0;
	if (SameSlope(l)) {
		return result;
	}

	if ( B == 0 ){
		result.x = -C/A;
		result.y = l.get_y(result.x);
		return result;
	}

	if ( l.B == 0){
		result.x = -l.C/l.A;
		result.y = get_y(result.x);
		return result;
	}

	result.x = (C*l.B - B*l.C)/(l.A*B - A*l.B);
	result.y = get_y(result.x);
	return result;
}  

/********************************************************************************/
bool Line::RayIntersection(const Ray& r, Vector &ppt) const{
	Line lRay(r);
	
	if (SameSlope(lRay))
		return false;
	
	ppt = intersection(lRay);
	return (r.InRightDir(ppt));
}

/********************************************************************************/
bool Line::IsPtCloserToPtOnLine(const Vector& point1, const Vector& point2, const Vector& target_point) const
{
	//if (!OnLine(target_point)) 
		//my_error("IsPtCloserToPtOnLine: target_point not on line");
  
	Vector p1 = ProjectPoint(point1);
	Vector p2 = ProjectPoint(point2);

	return p1.dist(target_point) < p2.dist(target_point);
}


/********************************************************************************/

//return true on top/left part of plane
bool Line::HalfPlaneTest(const Vector& point) const
{
	if (B==0)
		return point.x < -C/A;
	return point.y > get_y(point.x);
}

bool Line::SameSlope(const Line& l) const
{
	return B == 0 && l.B == 0 || A/B == l.A/l.B;
}

int Line::CircleIntersect(float radius, const Vector& center, Vector &psol1, Vector& psol2) const
{
	int numSol;
	psol1 = psol2 = Vector(0,0);
	if (fabs(A) > FLOAT_EPS) {
		float a,b,c;
		a = 1 + Sqr(B/A);
		b = 2*(-center.y + (C/A + center.x) * (B/A));
		c = -Sqr(radius) + Sqr(C/A + center.x) + Sqr(center.y);
		numSol = QuadraticFormula(a, b, c, psol1.y, psol2.y);
		psol1.x = - ( (B * psol1.y + C) / A );
		psol2.x = - ( (B * psol2.y + C) / A );
		return numSol;
	}
	else{
		numSol = QuadraticFormula(1, -2*center.x, Sqr(center.x) + Sqr(C/B + center.y) - Sqr(radius), psol1.x, psol2.x);
		psol1.y = psol2.y = - C / B;
		return numSol;
	}
}

/********************************************************************************/
/*                        Rect Class                                       */
/********************************************************************************/

Rect::Rect()
{
  left_x = right_x = top_y = bottom_y = 0;
}

/********************************************************************************/

Rect::Rect(float l, float r, float t, float b){
	if ( l >= r ) Swap(l, r);
	if ( t >= b ) Swap(t, b);

	left_x = l ; right_x = r ;
	top_y = t ; bottom_y = b ;
}
Rect::Rect(const Vector& center, const Vector& size){
	left_x = (float)(center.x - fabs(size.x)/2.0f);
	right_x = (float)(center.x + fabs(size.x)/2.0f) ;
	top_y = (float)(center.y - fabs(size.y)/2.0f);
	bottom_y = (float)(center.y + fabs(size.y)/2.0f);
}

bool Rect::IsWithin(const Vector &p) const
{
	return (bool) ((p.x >= left_x) && (p.x <= right_x) && (p.y >= top_y) && (p.y <= bottom_y)) ;
}

//order: TL, TR, BR, BL

Vector Rect::GetPoint(int n) const
{
	switch (n % 5) {
	case 0: return TopLeftCorner();
	case 1: return TopRightCorner();
	case 2: return BottomRightCorner();
	case 3: return BottomLeftCorner();
	case 4: return Center();
	}    
	//my_error("Rect::GetPoint: how did I get here");
	return Vector(0,0);
}

Vector Rect::Center() const
{
	return Vector((left_x+right_x)/2.0f, (bottom_y+top_y)/2.0f);
}

Line Rect::LeftEdge() const
{
	Line l;
	l.LineFromTwoPoints(TopLeftCorner(), BottomLeftCorner());
	return l;
}

Line Rect::RightEdge() const
{
	Line l;
	l.LineFromTwoPoints(TopRightCorner(), BottomRightCorner());
	return l;
}

Line Rect::TopEdge() const
{
	Line l;
	l.LineFromTwoPoints(TopLeftCorner(), TopRightCorner());
	return l;
}

Line Rect::BottomEdge() const
{
	Line l;
	l.LineFromTwoPoints(BottomLeftCorner(), BottomRightCorner());
	return l;
}

Line Rect::nearestHEdgeLine(const Vector& p) const
{
  return (fabs((p.y - top_y)) < fabs((bottom_y - p.y))) ? TopEdge() : BottomEdge() ;
}

Line Rect::nearestVEdgeLine(const Vector& p) const
{
  return (fabs((p.x - left_x)) < fabs((right_x - p.x))) ? LeftEdge() : RightEdge() ;
}

Line Rect::nearestEdgeLine(const Vector& p) const
{
  if(Min((float)fabs((p.x-left_x)),(float)fabs((right_x-p.x))) < Min((float)fabs((p.y-top_y)),(float)fabs((bottom_y-p.y))))
    return nearestVEdgeLine(p) ;
  else
    return nearestHEdgeLine(p) ;  
}

bool Rect::RayIntersection(const Ray& r,Vector &p) const
{
	Vector crosspoint;
	Vector returnp;
	float crossdist = -1;
	if (r.intersection(LeftEdge(),crosspoint))
	{
		if (crosspoint.y >=top_y && crosspoint.y <= bottom_y && (r.DistanceFromOrigin(crosspoint) < crossdist || crossdist == -1))
		{
			crossdist = r.DistanceFromOrigin(crosspoint);
			returnp = crosspoint;
		}
	}
	if (r.intersection(RightEdge(),crosspoint))
	{
		if (crosspoint.y >=top_y && crosspoint.y <=bottom_y && (r.DistanceFromOrigin(crosspoint) < crossdist || crossdist == -1))
		{
			crossdist = r.DistanceFromOrigin(crosspoint);
			returnp = crosspoint;
		}
	}
	if (r.intersection(TopEdge(),crosspoint))
	{
		if (crosspoint.x >= left_x && crosspoint.x <= right_x && (r.DistanceFromOrigin(crosspoint) < crossdist || crossdist == -1))
		{
			crossdist = r.DistanceFromOrigin(crosspoint);
			returnp = crosspoint;
		}
	}
	if (r.intersection(BottomEdge(),crosspoint))
	{
		if (crosspoint.x >=left_x && crosspoint.x <= right_x&& (r.DistanceFromOrigin(crosspoint) < crossdist || crossdist == -1))
		{
			crossdist = r.DistanceFromOrigin(crosspoint);
			returnp = crosspoint;
		}
	}
	if (crossdist  != -1)
	{
		p = returnp;
		return true;
	}
	else
	{
		p = Vector(0,0);
		return false;
	}
}


/*** The3rdCurve ********/
The3rdCurve::The3rdCurve(float a, float b, float c, float d, float inf, float sup){
	SetABCD(a, b, c, d);
	SetInfSup(inf, sup);
}

void The3rdCurve::SetABCD(float a, float b, float c, float d){
	this->a = a;
	this->b = b;
	this->c = c;
	this->d = d;
}

void The3rdCurve::SetInfSup(float inf, float sup){
	this->inf = inf;
	this->sup = sup;
}

float The3rdCurve::GetOutput(float x, bool limited) const
{
	float result = a *x * x *x + b * x * x + c * x + d;
	if (limited){
		pose_limitation(result, inf, sup);
	}
	return result;
}

bool The3rdCurve::Interpolation(float x0, float y0, float x1, float y1, float x2, float y2, float x3, float y3){
	if (x0 == x1 || x0 == x2 || x0 == x3 || x1 == x2 || x1 == x3 || x2 == x3) return false;
	float t0, t1, t2, t3;
	t0 = y0 / ((x0 - x1) * (x0 - x2) * (x0 - x3));
	t1 = y1 / ((x1 - x0) * (x1 - x2) * (x1 - x3));
	t2 = y2 / ((x2 - x0) * (x2 - x1) * (x2 - x3));
	t3 = y3 / ((x3 - x0) * (x3 - x1) * (x3 - x2));

	a = t0 + t1 + t2 + t3;
	b = - (t0 * (x1 + x2 + x3) + t1 * (x0 + x2 + x3) + t2 * (x0 + x1 + x3) + t3 * (x0 + x1 +x2));
	c = t0 * (x1 * x2 + x1 * x3 + x2 * x3) + t1 * (x0 * x2 + x0 * x3 + x2 * x3)
		+ t2 * (x1 * x0 + x3 * x1 + x3 * x0) + t3 * (x0 * x1 + x0 * x2 + x1 * x2);
	d = -(t0 * (x1 * x2 * x3) + t1 * (x0 * x2 * x3) + t2 * (x0 * x1 * x3) + t3 * (x0 * x1 * x2));
	return true;
}
/************************Ring Sector****************************/
//to determine whether a specific angle is in PossSpace angle section
bool RingSector::IsWithin(float angle){
	if(!valid) return false;

	int i;
	for(i=0; i<angles.Num_datas(); i++){
		if( AngleDifference(angles.Data(i).mean, angle) < angles.Data(i).delta){
			return true;
		}
	}
	return false;
}

///space became larger as time goes by
void RingSector::Diffuse(int time, float speed){
	if(!valid) return;

	radius.delta = Min(radius.mean , radius.delta + float(time*speed));

	int i;
	for(i=0; i<angles.Num_datas(); i++){
		angles.Data(i).delta += Rad2Deg(time*speed/radius.mean);
	}
	//check whether to merge
	for(i=angles.Num_datas()-1 ; i>0; i--){
		float dif = AngleDifference(angles.Data(i).mean, angles.Data(0).mean);
		if( dif < angles.Data(i).delta + angles.Data(0).delta){
			if(NormalizeAngle(angles.Data(i).mean - angles.Data(0).mean) > 0){
				angles.Data(0).mean = NormalizeAngle(angles.Data(0).mean +
					(angles.Data(i).delta - angles.Data(0).delta + dif)/2) ; 
			}else{
				angles.Data(0).mean = NormalizeAngle(angles.Data(i).mean +
					(angles.Data(0).delta - angles.Data(i).delta + dif)/2) ; 
			}
			angles.Data(0).delta += dif/2;
			angles.RmTail();
			continue;
		}
		break;
	}
}

void RingSector::MinusRadius(float inf, float sup){
	if(!valid) return;
	float n_sup = Min(radius.sup(), sup);
	float n_inf = Max(radius.inf(), inf);
	if(n_inf > n_sup){
		valid = false;
		return;
	}
	radius.mean = (n_sup + n_inf) / 2;
	radius.delta = (n_sup - n_inf) / 2;

}
//sight infor may diminish the RingSector
//Problem: when radius is too far and can't be seen, what it mean?
void RingSector::MinusAngle(NoiseData& subtrahend){
	if(!valid) return;

	int i;
	float default_angle = angles.Data(0).mean;
	for(i=angles.Num_datas()-1; i>=0; i--){
		float dif = AngleDifference(angles.Data(i).mean, subtrahend.mean);
		if(dif >= angles.Data(i).delta + subtrahend.delta) continue;
		if(dif + angles.Data(i).delta <= subtrahend.delta){
			if(angles.Num_datas()-1 > i){
				angles.Data(i) = angles.Data(angles.Num_datas()-1);
			}
			angles.RmTail();
			continue;
		}
		if( dif+subtrahend.delta <= angles.Data(i).delta){
			angles.Add(CreateAngleSector(
						subtrahend.mean + subtrahend.delta,
						angles.Data(i).mean + angles.Data(i).delta
						));
			angles.Data(i) = CreateAngleSector(
						angles.Data(i).mean - angles.Data(i).delta,
						subtrahend.mean - subtrahend.delta
						);
			continue;
		}
		if(NormalizeAngle(subtrahend.mean - angles.Data(i).mean) > 0){
			angles.Data(i) = CreateAngleSector(
				angles.Data(i).mean - angles.Data(i).delta,
				subtrahend.mean - subtrahend.delta
				);
		}else{
			angles.Data(i) = CreateAngleSector(
				subtrahend.mean + subtrahend.delta,
				angles.Data(i).mean + angles.Data(i).delta
			);
		}
	}

	if(angles.Num_datas() == 0){
		valid = false;
	}else if(angles.Num_datas() > 1){		
		if(angles.Data(0).delta < angles.Data(1).delta){
			Swap(angles.Data(0), angles.Data(1));
		}
	}
}

void RingSector::Intersect(NoiseData r, NoiseData a){
	if(!valid) return;

	//intersect radius range
	float dif_r = (float)fabs(r.mean - radius.mean);
	if(dif_r >= r.delta + radius.delta ){
		angles.Reset();
		valid = false;
		return;
	}
	radius.Intersect(r);
	float f;
	int i;
	for(i=angles.Num_datas()-1; i>=0; i--){
		f=a.mean - angles.Data(i).mean;
		if(f < -180.0f) a.mean += 360.0f;
		else if (f > 180.0f) angles.Data(i).mean += 360.0f;
		if(!angles.Data(i).Intersect(a)){
			angles.Data(i) = angles.Data(angles.Num_datas()-1);
			angles.RmTail();
		}else{
			angles.Data(i).mean = NormalizeAngle(angles.Data(i).mean);
		}
	}
	if(angles.Num_datas() == 0){
		valid = false;
	}else if(angles.Num_datas() > 1){
		if(angles.Data(0).delta < angles.Data(1).delta){
			Swap(angles.Data(0), angles.Data(1));
		}
	} 
}

//Note: only the main angle range is consideded
void RingSector::ShiftCenter(const Vector& delta){
	Vector v =  Polar2Vector(radius.mean, angles.Data(0).mean) - delta;
	radius.mean = v.mod();
	angles.Data(0).mean = v.Angle();
}

///calculate total area  
float RingSector::GetArea(){
	float sum = 0;
	int i;
	for(i=0; i<angles.Num_datas(); i++){
		sum += angles.Data(i).delta;
	} 
	sum = 4 * Deg2Rad(sum) * radius.mean * radius.delta;
	return sum;
}

Vector RingSector::GetCenterPos(){
	Vector cpos;
	if(valid){
		cpos = Polar2Vector(radius.mean, angles.Data(0).mean);
	}
	return cpos;
}

void RingSector::operator =(const RingSector &ps){
	valid = ps.IsValid();
	if(!valid)	return;
	radius = ps.radius;
	angles = ps.angles;
}

//Polar2Vector(angle, dist)
float RingSector::Dist(float angle, float dist){
	float Min_dist = 1000.0f;
	if(!valid) return Min_dist;
	float dif, tmpdif, tmp_dist;
	int i;
	for(i=0; i<angles.Num_datas(); i++){
		dif = AngleDifference(angles.Data(i).mean, angle); 
		tmpdif = dif - angles.Data(i).delta;
		if(tmpdif <= 0){
			if(dist < radius.inf()){
				tmp_dist = radius.inf() - dist;
			}else if(dist > radius.sup()){
				tmp_dist = dist - radius.sup();
			}else
				return 0.0f;
		}else{
			float p = dist * Cos(tmpdif);
			if(p < radius.inf()){
				tmp_dist = (float)sqrt( Sqr(radius.inf()-p) + Sqr(dist*Sin(tmpdif)) );
			}else if(p > radius.sup()){
				tmp_dist = (float)sqrt( Sqr(p-radius.sup()) + Sqr(dist*Sin(tmpdif)) );
			}else
				tmp_dist = dist * Sin(tmpdif);
		}
		Min_dist = Min(Min_dist, tmp_dist);
	}
	return Min_dist;
}

bool RingSector::Isthrough(float angle, float dist){
	if(!valid) return false;
	if(dist < radius.inf()) return false;
	int i;
	for(i=0; i<angles.Num_datas(); i++){
		float dif = AngleDifference(angles.Data(i).mean, angle); 
		if(dif <= angles.Data(i).delta) return true;
	}
	return false;
}

float RingSector::Getmax_x(){
	if(!valid) return 0;
	float min_angle = 180.0;
	int i;
	for(i=0; i<angles.Num_datas(); i++){
		if(NormalizeAngle(0 - angles.Data(i).inf(), 0) < 2 * angles.Data(i).delta){
			min_angle = 0;
			break;
		}
		float tmp0 = (float)fabs(NormalizeAngle(angles.Data(i).inf()));
		float tmp1 = (float)fabs(NormalizeAngle(angles.Data(i).sup()));
		min_angle = Min(min_angle, Min(tmp0, tmp1));
	} 
	if(min_angle <= 90){
		return radius.sup() * Cos(min_angle); 
	}else{
		return radius.inf() * Cos(min_angle);
	}
}

float RingSector::Getmin_x(){
	if(!valid) return 0;
	float max_angle = 0;
	int i;
	for(i=0; i<angles.Num_datas(); i++){
		if(NormalizeAngle(180 - angles.Data(i).inf(), 0) < 2 * angles.Data(i).delta){
			max_angle = 180.0f;
			break;
		}
		
		float tmp0 = (float)fabs(NormalizeAngle(angles.Data(i).inf()));
		float tmp1 = (float)fabs(NormalizeAngle(angles.Data(i).sup()));
		max_angle = Max(max_angle, Max(tmp0, tmp1));
	} 
	if(max_angle <= 90){
		return radius.inf() * Cos(max_angle); 
	}else{
		return radius.sup() * Cos(max_angle);
	}
}

bool RingSector::IsWithin(float angle, float dist)
{
	return IsWithin(angle) && radius.IsWithin(dist);
}

